/*
 * Copyright (c) 2008, 2009, Bjoern B. Brandenburg <bbb [at] cs.unc.edu>
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/* Starvation-free, real-time-friendly reader-writer locks.
 *
 * This is an implementation of phase-fair reader-writer locks for SPARC
 * V9. Readers are only blocked by at most one reader and one writer
 * phase. Writers gain access in FIFO order. Both reader and writer paths
 * require two atomic CAS instructions (each with a retry loop) to emulate
 * atomic addition and fetch-and-add.
 *
 * For details, see:
 *
 *     B. Brandenburg and J. Anderson, "Reader-Writer Synchronization for
 *     Shared-Memory Multiprocessor Real-Time Systems", Proceedings of the 21th
 *     Euromicro Conference on Real-Time Systems (ECRTS 2009),
 *     pp. 184-193. IEEE, July 2009.
 *
 * Tested on a SUN UltraSPARC T1 ("Niagara") under Linux with gcc 4.1.3.
 */

#ifndef PFRW_H
#define PFRW_H

typedef struct {
	u32 rin, win, rout, wout;
} pf_rwlock_t;

#define DEFINE_PF_RWLOCK(x) pf_rwlock_t x =	\
	{ 0, 0, 0, 0 }

static inline void pf_read_lock(pf_rwlock_t *l)
{
	unsigned long tmp1, tmp2;

	__asm__ __volatile__(
"1:	lduw		[%2], %0	\n"
"	add		%0, 256, %1	\n"
"	cas		[%2], %0, %1	\n"
"	cmp		%0, %1		\n"
"	membar		#StoreLoad | #StoreStore \n"
"	bne,pn       	%%icc, 1b	\n"
"	 andcc		%1, 0xff, %1	\n"
"	bnz,a,pn	%%icc, 2f	\n"
"	 nop				\n"
"	.subsection	2		\n"
"2:	lduw		[%2], %0	\n"
"3:	membar		#LoadLoad	\n"
"	and		%0, 0xff, %0	\n"
"	cmp		%0, %1		\n"
"	bne,pt		%%icc, 4f	\n"
"	 lduw		[%2], %0	\n"
"	ba,a,pt		%%icc, 3b	\n"
"	.previous			\n"
"4:"
	: "=&r" (tmp1), "=&r" (tmp2)
	: "r" (&l->rin)
	: "memory");
}

static inline void pf_read_unlock(pf_rwlock_t *l)
{
	unsigned long tmp1, tmp2;

	__asm__ __volatile__(
"	membar		#StoreLoad | #LoadLoad \n"
"1:	lduw		[%2], %0	\n"
"	add		%0, 256, %1	\n"
"	cas		[%2], %0, %1	\n"
"	cmp		%0, %1		\n"
"	bne,pn       	%%icc, 1b	\n"
"	 nop				\n"
	: "=&r" (tmp1), "=&r" (tmp2)
	: "r" (&l->rout)
	: "memory");
}

static inline void pf_write_lock(pf_rwlock_t *l)
{
	unsigned long tmp1, tmp2, tmp3;

	__asm__ __volatile__(
"1:	lduw		[%3], %0	\n"
"	add		%0, 0x2, %1	\n"
"	cas		[%3], %0, %1	\n"
"	cmp		%0, %1		\n"
"	bne,pn       	%%icc, 1b	\n"
"	 membar		#StoreLoad | #StoreStore \n"
"2:	lduw		[%4], %2	\n"
"	cmp		%0, %2		\n"
"	bne,pn		%%icc, 2b	\n"
"	 membar		#LoadLoad	\n"
"	and		%0, 0xff, %0	\n"
"	add		%0, 0x1, %0	\n"
"3:	lduw		[%5], %1	\n"
"	add		%0, %1, %2	\n"
"	cas		[%5], %1, %2	\n"
"	cmp		%1, %2		\n"
"	bne,pn       	%%icc, 3b	\n"
"	 andn		%1, 0xff, %1	\n"
"4:	lduw		[%6], %2	\n"
"	cmp		%1, %2		\n"
"	bne,pn       	%%icc, 4b	\n"
"	 membar		#LoadLoad	\n"
	: "=&r" (tmp1), "=&r" (tmp2), "=&r" (tmp3)
	: "r" (&l->win), "r" (&l->wout), "r" (&l->rin), "r" (&l->rout)
	: "memory");

}

static inline void pf_write_unlock(pf_rwlock_t *l)
{
	unsigned long tmp;
	__asm__ __volatile__(
"	membar		#LoadStore | #StoreStore \n"
"	stb		%%g0, [%1+0x3]	\n"
"	membar		#StoreStore	\n"
"	lduw		[%2], %0	\n"
"	add		%0, 0x2, %0	\n"
"	stw		%0, [%2]	\n"
	: "=&r" (tmp)
	: "r" (&l->rin), "r" (&l->wout)
	: "memory");
}

#endif
