#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <pthread.h>

#include <unistd.h>
#include <linux/unistd.h>
#include <sys/types.h>
#include <sched.h>

#include "parallel.h"

#include "kernel_compat.h"
#include "mem.h"
#include "atomic.h"
#include "cycles.h"

#include "locks/spin.h"
#include "locks/ticket.h"
#include "locks/ticket_proto1.h"
#include "locks/ticket_proto2.h"
#ifndef __i386__
#include "locks/mcs.h"
#endif
#include "locks/mcs_proto.h"


#include "locks/barw_proto1.h"
#include "locks/barw_proto2.h"
#include "locks/pfrw.h"
#include "locks/linrw.h"
#include "locks/mcsrw_proto.h"
#ifndef __i386__
#include "locks/mcsrw.h"
#include "locks/mcssrw.h"
#endif
#include "locks/mcssrw_proto.h"

#ifdef __i386__
#include <sys/io.h>
#include "irq.h"
#else
#define cli()
#define sti()
#endif

#define WARMUP    100
#define LOOPS 3000000

#define BENCH(name, def)			\
	void bench_ ## name (int loops)		\
	{					\
		cycles_t start, stop;		\
		cycles_t t0, t1, sum = 0, max = 0;	\
		int i;				\
		def(s);				\
		for (i = 0; i < WARMUP; i++) {	\
			name ## _lock(&s);		\
			cpu_relax();		\
			name ## _unlock(&s);	\
		}				\
		start = get_cycles();		\
		for (i = 0; i < loops; i++) {	\
			name ## _lock(&s);		\
			cpu_relax();		\
			name ## _unlock(&s);	\
		}				\
		stop = get_cycles();					\
		for (i = 0; i < loops; i++) {				\
			cli();						\
			t0 = get_cycles();	\
			name ## _lock(&s);	\
			cpu_relax();		\
			name ## _unlock(&s);	\
                        t1 = get_cycles();      \
			sti();			\
			t1 -= t0;		\
			if (t1 > max)		\
				max = t1;	\
			sum += t1;		\
		}				\
		printf("\t%-20s: %9.5fcy %9.5fcy %9dcy\n",  		\
		       #name, (stop - start) / (double) loops,		\
		       (sum) / (double) loops, (int) (max)		\
			);			\
	}

#define BENCHQ(name, def, qdef)			\
	void bench_ ## name (int loops)		\
	{					\
		cycles_t start, stop;  		\
		cycles_t t0, t1, sum = 0, max = 0;\
		int i;				\
		def(s);				\
		qdef(x);			\
		for (i = 0; i < WARMUP; i++) {	\
			name ## _lock(&s, &x);	\
			cpu_relax();		\
			name ## _unlock(&s, &x);	\
		}				\
		start = get_cycles();		\
		for (i = 0; i < loops; i++) {	\
			name ## _lock(&s, &x);	\
			cpu_relax();		\
			name ## _unlock(&s, &x);	\
		}				\
		stop = get_cycles();					\
		for (i = 0; i < loops; i++) {	\
			cli();			\
			t0 = get_cycles();	\
			name ## _lock(&s, &x);	\
			cpu_relax();		\
			name ## _unlock(&s, &x);	\
                        t1 = get_cycles();      \
			sti();			\
			t1 -= t0;		\
			if (t1 > max)		\
				max = t1;	\
			sum += t1;		\
		}				\
		printf("\t%-20s: %9.5fcy %9.5fcy %9dcy\n",  		\
		       #name, (stop - start) /  (double) loops,		\
		       (sum) / (double) loops, (int) (max)		\
			);			\
	}


BENCH(spin, DEFINE_SPINLOCK);
BENCH(ticket_proto1, DEFINE_PROTO1LOCK);
BENCH(ticket_proto2, DEFINE_PROTO2LOCK);
BENCH(ticket, DEFINE_TICKETLOCK);
BENCH(ba_proto1_read, DEFINE_BA_PROTO1_RWLOCK);
BENCH(ba_proto1_write, DEFINE_BA_PROTO1_RWLOCK);
BENCH(ba_proto2_read, DEFINE_BA_PROTO2_RWLOCK);
BENCH(ba_proto2_write, DEFINE_BA_PROTO2_RWLOCK);
BENCH(pf_write, DEFINE_PF_RWLOCK);
BENCH(pf_read, DEFINE_PF_RWLOCK);
BENCH(lin_read, DEFINE_LIN_RWLOCK);
BENCH(lin_write, DEFINE_LIN_RWLOCK);
BENCHQ(mcsp, DEFINE_MCSPLOCK, DEFINE_MCSPNODE);
BENCHQ(mcsp_write, DEFINE_MCS_RWLOCK, DEFINE_MCS_RWNODE);
BENCHQ(mcsp_read,  DEFINE_MCS_RWLOCK, DEFINE_MCS_RWNODE);
#ifndef __i386__
BENCHQ(mcs, DEFINE_MCSLOCK, DEFINE_MCSNODE);
BENCHQ(mcs_write, DEFINE_MCS_RWLOCK, DEFINE_MCS_RWNODE);
BENCHQ(mcs_read,  DEFINE_MCS_RWLOCK, DEFINE_MCS_RWNODE);
BENCH(mcss_read, DEFINE_MCSS_RWLOCK);
BENCH(mcss_write, DEFINE_MCSS_RWLOCK);
#endif
BENCH(mcssp_read, DEFINE_MCSSP_RWLOCK);
BENCH(mcssp_write, DEFINE_MCSSP_RWLOCK);

void show(pf_rwlock_t *lock, int x)
{
	printf("  rin:0x%x rout:0x%x win:%d wout:%d -> %d\n",
	       lock->rin, lock->rout, lock->win, lock->wout, x);
}

void test(void)
{
/* trylock is missing */
#ifndef __sparc__
	int x;
	DEFINE_PF_RWLOCK(lock);

	lock.rin  = 1024;
	lock.rout = 1024;
	lock.win  = 16;
	lock.wout = 16;
	printf("attempting...\n");
	show(&lock, 0);
	x = pf_write_trylock(&lock);
	show(&lock, x);
	x = pf_read_trylock(&lock);
	show(&lock, x);
	x = pf_write_trylock(&lock);
#endif
}

void bench(int loops)
{
	printf("\t%-20s: %9s   %9s   %9s\n", "Lock Op", "total avg", "ind. avg", "max");
	printf("\t%-20s: %9s   %9s   %9s\n", "=======", "=========", "========", "===");

	printf("Mutex:\n");
	bench_spin(loops);
	mb();
	bench_ticket_proto1(loops);
	mb();
	bench_ticket_proto2(loops);
	mb();
	bench_ticket(loops);
	mb();
	bench_mcsp(loops);
	mb();
#ifndef __i386__
	bench_mcs(loops);
	mb();
#endif
	printf("RW:\n");
	bench_ba_proto1_read(loops);
	mb();
	bench_ba_proto2_read(loops);
	mb();
	bench_pf_read(loops);
	mb();
	bench_ba_proto1_write(loops);
	mb();
	bench_ba_proto2_write(loops);
	mb();
	bench_pf_write(loops);
	mb();
	bench_lin_read(loops);
	mb();
	bench_lin_write(loops);
	mb();
	bench_mcsp_read(loops);
	mb();
	bench_mcsp_write(loops);
	mb();
#ifndef __i386__
	bench_mcs_read(loops);
	mb();
	bench_mcs_write(loops);
	mb();
	bench_mcss_read(loops);
	mb();
	bench_mcss_write(loops);
#endif
	mb();
	bench_mcssp_read(loops);
	mb();
	bench_mcssp_write(loops);
	printf("---\n");
}


int main(int argc, char** argv)
{
	srand(time(NULL));

#ifdef __i386__
	if (geteuid() != 0) {
		fprintf(stderr, "Need root privileges.\n");
		exit(1);
	}
	iopl(3);
#endif

	bench(LOOPS);
	bench(LOOPS);
	bench(LOOPS);
	return 0;
}
