#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <time.h>
#include <string.h>

#include "parallel.h"
#include "units.h"
#include "cycles.h"
#include "locks/rwlocks.h"

#define streq(a, b) (strcmp(a, b) == 0)

DEF_RW_LOCK(global_lock);
int counter = 0;
int last_id = -1;
int repetition_counter = 0;

#define LOOPS_PER_THREAD 200000

#ifdef HAVE_RW_READER_COUNT
#undef HAVE_RW_READER_COUNT
#endif

float write_ratio = 0.5;
float delay_ratio = 1.0;

int num_threads = 0;

cycles_t min[MAX_THREADS],
	max[MAX_THREADS],
	total[MAX_THREADS],
	exec[MAX_THREADS],
	total_exec[MAX_THREADS];
double cs_ratio[MAX_THREADS];
int local_rep[MAX_THREADS];

atomic_t writes, reads;

void init(void)
{
	int i;
	counter = 0;
	for (i = 0; i < num_threads; i++) {
		atomic_set(&writes, 0);
		atomic_set(&reads, 0);
		min[i] = ULONG_MAX;
		max[i] = 0;
		total[i] = 0;
		local_rep[i] = 0;
		exec[i] = 0;
	}
}

void show_thread_info(void)
{
	int i;
	for (i = 0; i < num_threads; i++) {
		printf("[%2d] csr:%f exec:%f total_exec:%f\n",
		       i,
		       exec[i] /  (double) total_exec[i],
		       cycles2usec(exec[i]),
		       cycles2usec(total_exec[i]));
	}
}

void show_hdr(void)
{
	printf("%3s, %6s, %6s, "
	       "%10s, %10s, %10s, %10s, %10s, %14s\n",
	       "m", "%w", "%cs", "blk del/m", "sblk del/m", "ex del",
	       "max ex / m", "max_exe", "sum ex / m");
}

void show_csv(void)
{
	int i;
	int max_blk, min_blk, max_total_blk, min_total_blk, max_exe, min_exe;
	double sum_exe = 0;
	double avg_csr = 0;
	max_blk = min_blk = max_total_blk = min_total_blk = max_exe = min_exe = 0;
	for (i = 0; i < num_threads; i++) {
		if (max[i] > max[max_blk])
			max_blk = i;
		if (max[i] < max[min_blk])
			min_blk = i;
		if (total[i] > total[max_total_blk])
			max_total_blk = i;
		if (total[i] < total[min_total_blk])
			min_total_blk = i;
		if (exec[i] < exec[min_exe])
			min_exe = i;
		if (exec[i] > exec[max_exe])
			max_exe = i;
		sum_exe += cycles2usec(exec[i]);
		avg_csr += exec[i] /  (double) total_exec[i];
	}
	avg_csr /= num_threads;
	printf("%3d, %6.4f, %6.4f, "
	       "%10.4f, %10.4f, %10.4f, %10.4f, %10.4f, %14.4f\n",
	       num_threads,
	       atomic_read(&writes) / ((double) num_threads * LOOPS_PER_THREAD),
	       avg_csr,
	       cycles2usec(max[max_blk] -  max[min_blk]) / num_threads,
	       cycles2usec((total[max_total_blk] - total[min_total_blk])
			   / 1000) / num_threads,
	       cycles2usec((exec[max_exe] - exec[min_exe])
			   / 1000) / num_threads,
	       cycles2usec(exec[max_exe]) / (num_threads * LOOPS_PER_THREAD),
	       cycles2usec(exec[max_exe]) / 1000,
	       sum_exe / num_threads);
}

void wait(cycles_t cycles)
{
	cycles_t now, when;
	when = get_cycles() + cycles;
	do {
		now = get_cycles();
	} while (now < when);
//	printf("waited for %lu\n", cycles);
}

atomic_t seqno = ATOMIC_INIT(0);

#ifdef HAVE_RW_LOCK_CTX
DEF_RW_LOCK_CTX(*ctx_ref[MAX_THREADS]);
#endif

int where_at[MAX_THREADS];
cycles_t last_start[MAX_THREADS];

void bang_on_lock(int id, void* _)
{
	unsigned int seed = get_cycles() / gettid();
	int i, c, w, on_entry;
#ifdef HAVE_RW_READER_COUNT
	u32 num_readers = 0;
#endif
#ifdef USE_NONE_RWLOCK
	int diffctr = 0;
#endif
	int do_write;
	cycles_t start, stop, t1, t2, bang_start;
	DEF_RW_LOCK_CTX(lock_ctx);

//	printf("[%d] active\n", id);

#ifdef HAVE_RW_LOCK_CTX
	ctx_ref[id] = &lock_ctx;
#endif
	bang_start = get_cycles();
	for (i = 0; i < LOOPS_PER_THREAD; i++) {
		where_at[id] = 1;
//		printf("[%d] doing %d\n", id, i);
		t1 = get_cycles();
		do_write = rand_r(&seed) < RAND_MAX * write_ratio;
//		last_start[id] = get_cycles();
		start = get_cycles();
//		pre = atomic_inc_return(&seqno);
		if (do_write) {
			write_lock(&global_lock, lock_ctx);
#ifdef HAVE_RW_READER_COUNT
			num_readers = reader_count(&global_lock);
#endif
		} else
			read_lock(&global_lock, lock_ctx);
		stop  = get_cycles();
		on_entry = atomic_read(do_write ?  &reads  : &writes);
//		where_at[id] = 2;
//		enter = atomic_inc_return(&seqno);
		stop -= start;
		atomic_inc(do_write ? &writes : &reads);
		if (stop < min[id])
			min[id] = stop;
		if (stop > max[id])
			max[id] = stop;
		total[id] += stop;
		if (do_write) {
			/* writer */
			c = counter++;
			repetition_counter += last_id == id;
			local_rep[id] += last_id == id;
			last_id = id;
//			exit = atomic_inc_return(&seqno);
#ifdef HAVE_RW_READER_COUNT
			if (num_readers)
				printf("[%2d] BAD: reader_count=%u\n",
				       id, num_readers);
#endif
#ifndef USE_NONE_RWLOCK
			w = atomic_read(&reads);
			if (on_entry != w)
				printf("[%2d] BAD: on_entry:%d != r:%d\n",
				       id, on_entry, w);
#endif
			write_unlock(&global_lock, lock_ctx);
		} else {
			/* read */
			c = counter;
			w = atomic_read(&writes);
			if (c != w)
#ifndef USE_NONE_RWLOCK
				printf("[%2d] BAD: c:%d != w:%d\n", id, c, w);
#else
			diffctr++;
#endif
#ifndef USE_NONE_RWLOCK
			w = atomic_read(&writes);
			if (on_entry != w)
				printf("[%2d] BAD: on_entry:%d != w:%d\n",
				       id, on_entry, w);
#endif

//			exit = atomic_inc_return(&seqno);
			read_unlock(&global_lock, lock_ctx);
		}
		t2 = get_cycles();
		exec[id] += t2 - t1;
//		where_at[id] = 3;
//		printf("delta: %lu, %f, delay: %lu\n", t2 - t1, delay_ratio, (cycles_t) ((t2 - t1) * delay_ratio));
		wait((t2 - t1) * delay_ratio);
/*		printf("%d [2%d] pre\n", pre, id);
		printf("%d [%2d] enter  c=%d i=%dw=%d\n", enter, id, c, i, do_write);
		printf("%d [%2d] exit\n", exit, id);
*/	}
	total_exec[id] = get_cycles() - bang_start;
//	printf("[%d] done.\n", id);
}

static void do_perf_exp(int min_threads, int show_ti)
{
	int j;
	int warmup =  1;
	int sample =  2;

	show_hdr();
	for (num_threads = min_threads; num_threads <= MAX_THREADS; num_threads++) {
		for (j = 0; j < warmup; j++) {
			init();
			in_parallel(num_threads, bang_on_lock, NULL, 1, 1);
		}
		for (j = 0; j < sample; j++) {
			init();
			in_parallel(num_threads, bang_on_lock, NULL, 1, 1);
			show_csv();
			fflush(stdout);
			if (show_ti)
				show_thread_info();
		}

	}
}

static u32 rcount[MAX_THREADS], wcount[MAX_THREADS];

static void cinit(void)
{
	int i;
	for (i = 0; i < MAX_THREADS; i++)
	{
		rcount[i] = 0;
		wcount[i] = 0;
	}
	atomic_set(&reads, 0);
	atomic_set(&writes, 0);
}

void count_lock_acq(int id, void* _)
{
	unsigned int seed = get_cycles() / gettid();
	int do_write;
	cycles_t start, stop;
	DEF_RW_LOCK_CTX(lock_ctx);

#ifdef HAVE_RW_LOCK_CTX
	ctx_ref[id] = &lock_ctx;
#endif
	start = get_cycles();
	stop  = start + usec2cycles(1000 * 1000); /* one second in usec */
	while (get_cycles() < stop) {
		do_write = rand_r(&seed) < RAND_MAX * write_ratio;
		if (do_write)
			write_lock(&global_lock, lock_ctx);
		else
			read_lock(&global_lock, lock_ctx);
		mb();
		if (do_write) {
			write_unlock(&global_lock, lock_ctx);
			wcount[id]++;
		} else {
			read_unlock(&global_lock, lock_ctx);
			rcount[id]++;
		}
	}
}

static void show_count_csv()
{
	u32 rtotal = 0, wtotal = 0;
	int i;
	for (i = 0; i < MAX_THREADS; i++)
	{
		rtotal += rcount[i];
		wtotal += wcount[i];
	}
	printf("%3d, %6.4f, %6.4f, %10u, %10u, %10u\n",
	       num_threads,
	       write_ratio,
	       wtotal / (double) (wtotal + rtotal),
	       rtotal,
	       wtotal,
	       rtotal + wtotal);
}

static void do_count_exp(int min_threads, int show_ti)
{
	int j;
	int warmup =  1;
	int sample =  2;

	for (num_threads = min_threads; num_threads <= MAX_THREADS; num_threads++) {
		for (j = 0; j < warmup; j++) {
			cinit();
			in_parallel(num_threads, count_lock_acq, NULL, 1, 1);
		}
		for (j = 0; j < sample; j++) {
			cinit();
			in_parallel(num_threads, count_lock_acq, NULL, 1, 1);
			show_count_csv();
			fflush(stdout);
//			if (show_ti)
//				show_thread_info();
		}
	}
}


static void usage(const char* msg)
{
        fprintf(stderr, "Error: %s\n", msg);
        exit(1);
}

#define OPTSTR "w:d:tn:e:"

int main(int argc, char** argv)
{
	int opt;
	int show_ti = 0;
	int min_threads = 2;
	enum {
		PERF,
		COUNT,
	} todo = PERF;

	while ((opt = getopt(argc, argv, OPTSTR)) != -1) {
		switch (opt) {
		case 'w':
			if (!str2float(optarg, &write_ratio))
				usage("bad argument");
			if (0 > write_ratio || 1.0 < write_ratio)
				usage("write ratio must be in [0, 1.0]");
			break;
		case 'd':
			if (!str2float(optarg, &delay_ratio))
				usage("bad argument");
			if (0 > delay_ratio)
				usage("delay ratio < 0");
			break;
		case 't':
			show_ti = 1;
			break;
		case 'n':
			if (!str2int(optarg, &min_threads))
				usage("bad argument");
			if (0 >= min_threads)
				usage("min_threads <= 0");
			break;
		case 'e':
			if (streq(optarg, "perf"))
				todo = PERF;
			else if (streq(optarg, "count"))
				todo = COUNT;
			else
				usage("-e [perf|count] only");
			break;
		case ':':
			usage("Argument missing.");
			break;
		case '?':
		default:
			usage("Bad argument.");
			break;
		}
	}
	switch (todo) {
	case PERF:
		do_perf_exp(min_threads, show_ti);
		break;
	case COUNT:
		do_count_exp(min_threads, show_ti);
		break;
	}
	return 0;
}
